#include <Servo.h>
#include <SoftwareSerial.h>
#include <Wire.h>
#include <RTClib.h>
#include <time.h>
#include <hd44780.h>                       // main hd44780 header
#include <hd44780ioClass/hd44780_I2Cexp.h> // i2c expander i/o class header
#include "HX711.h"

#define IDLE 0
#define FEED_TIME 1
#define SET_FEEDING_TIME 2
#define NO_FEED_SET 3

#define rxPin 10
#define txPin 11
SoftwareSerial bt = SoftwareSerial(rxPin, txPin);

Servo servo;
int servoPin = 9;

#define SCALE_OFFSET 218
const uint8_t dataPin = A1;
const uint8_t clockPin = A0;
HX711 scale;

#define CLOCK_INTERRUPT_PIN 2
RTC_DS3231 rtc;

hd44780_I2Cexp lcd(0x27);

static int state = NO_FEED_SET;

int feeding_hour;
int feeding_mins;
DateTime feeding_time, now;
time_t now_sec, feeding_hour_sec;
int remaining_mins, remaining_sec;

const int BUFFER_SIZE = 10;
char buf[BUFFER_SIZE];
char str[BUFFER_SIZE];

// ------------------ SERVO ----------------
void setup_servo() {
  servo.attach(servoPin);
  servo.write(0);
}

void rotate_servo() {
  for (int pos = 0; pos <= 90; pos += 5) {
    servo.write(pos);
    delay(15);
  }

  delay(450);

  for (int pos = 90; pos >= 0; pos -= 5) {
    servo.write(pos);
    delay(15);
  }

  delay(1000);
}

// ---------------- SCALE ----------------
void scale_setup() {
  scale.begin(dataPin, clockPin);
  // scale.tare();
  scale.callibrate_scale(SCALE_OFFSET, 5);
}

float get_weight() {
  float weight = SCALE_OFFSET - scale.get_units(20);
  if (weight < 0) {
    weight = 0;
  }

  return weight;
}

void send_weight() {
  bt.print("Current food weight is ");
  bt.println(get_weight());
}

// ---------------- BLUETOOTH ---------------
// bluetooth communication
void bt_setup()  {
    // Define pin modes for TX and RX
    pinMode(rxPin, INPUT);
    pinMode(txPin, OUTPUT);
    
    // Set the baud rate for the SoftwareSerial object
    bt.begin(9600);
}

// get command via bluetooth
void get_bt_command() {
  int i = 0;
  while (bt.available() > 0) {
      char command = bt.read();
      str[i++] = command;
      if (command == '\n') {
        break;
      }
  }
  if (i != 0) {
    if (str[0] == 'T') {
      state = SET_FEEDING_TIME;
    }
    if (str[0] == 'F') {
      state = FEED_TIME;
    }
    if (str[0] == 'W') {
      send_weight();
    }
  }
}

// get feeding time via bluetooth
// format has to be hours:mins
void bt_get_feeding_time() {
  bt.println("Set feed time!");
  bt.println("Format hour:mins");
  lcd.clear();
  lcd.setCursor(0, 1);
  lcd.print("Time format h:m");

  while(bt.available() == 0) {
  }
  bt.readBytes(buf, 6);
  buf[5] = '\0';
  str[0] = buf[0];
  str[1] = buf[1];
  str[2] = '\0';
  feeding_hour = atoi(str);
  str[0] = buf[3];
  str[1] = buf[4];
  str[2] = '\0';
  feeding_mins = atoi(str);

  digitalWrite(rxPin, HIGH);   // deactivate the RFID reader for a moment so it will not flood
  bt.flush();                  // clear the buffer
  delay(1500);

  bt.print("Feed time set to ");
  bt.print(feeding_hour);
  bt.print(":");
  bt.println(feeding_mins);

  lcd.clear();
  lcd.setCursor(0, 1);
  lcd.print("Next feed ");
  lcd.print(feeding_hour);
  lcd.print(":");
  lcd.print(feeding_mins);

  delay(5000);

  digitalWrite(rxPin, LOW);

  lcd.clear();
}

// ------------------- LCD ---------------------
void lcd_setup() {
  lcd.begin(16,2);  // initialize the lcd 
  lcd.backlight();//To Power ON the back light
}

void print_next_feed() {
  now_sec = rtc.now().unixtime();

  lcd.clear();
  lcd.setCursor(0, 1);
  lcd.print("Feed in");
  lcd.print(" ");
  remaining_mins = (int)((feeding_hour_sec - now_sec) / 60.0);
  lcd.print(remaining_mins);
  lcd.print(':');
  remaining_sec = (int)((feeding_hour_sec - now_sec) % 60);
  lcd.print(remaining_sec);
  delay(1000);
}

// ------------------- RTC ---------------------
void onAlarm() {
  // Set status flag
  state = FEED_TIME;
}

void set_alarm() {
  // set alarm 1, 2 flag to false (so alarm 1, 2 didn't happen so far)
  // if not done, this easily leads to problems, as both register aren't reset on reboot/recompile
  rtc.clearAlarm(1);
  rtc.clearAlarm(2);

  // stop oscillating signals at SQW Pin
  // otherwise setAlarm1 will fail
  rtc.writeSqwPinMode(DS3231_OFF);

  // turn off alarm 2
  rtc.disableAlarm(2);

  // get feeding time for alarm
  now = rtc.now();
  feeding_time = DateTime(now.year(), now.month(), now.day(), feeding_hour, feeding_mins, 0);
  feeding_hour_sec = feeding_time.unixtime();

  // schedule an alarm
  rtc.setAlarm1(feeding_time, DS3231_A1_Hour);
}

void rtc_setup() {
  if (!rtc.begin()) {
    lcd.print("Couldn't find RTC");
    while (1);
  }

  rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));

  //we don't need the 32K Pin, so disable it
  rtc.disable32K();

  // Making it so, that the alarm will trigger an interrupt
  pinMode(CLOCK_INTERRUPT_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(CLOCK_INTERRUPT_PIN), onAlarm, FALLING);
}

void setup() {
  scale_setup();
  lcd_setup();
  setup_servo();
  bt_setup();
  rtc_setup();

  bt.println("Pet feeder up and running!");
  lcd.setCursor(0, 1);
  lcd.print("Pet feeder up!");
  delay(1000);
}

void loop() {
  if (state == NO_FEED_SET) {
    lcd.clear();
    lcd.setCursor(0, 1);
    lcd.print("Feed not set!");
    delay(500);
    get_bt_command();
  } else {
    if (state == SET_FEEDING_TIME) {
      bt_get_feeding_time();
      set_alarm();
      state = IDLE;
    } else {
      if (state == FEED_TIME) {
        rtc.clearAlarm(1);
        bt.println("Feeding pet now!");

        lcd.clear();
        lcd.setCursor(0, 1);
        lcd.print("Feeding now!");

        rotate_servo();
        delay(1000);
        bt.println("Feeding done!");
        send_weight();
        state = NO_FEED_SET;
      } else {
        if (state == IDLE) {
          print_next_feed();
          get_bt_command();
        }
      }
    }
  }
}
